/*
 * include/linux/dyn_pageflags.h
 *
 * Copyright (C) 2004-2005 Nigel Cunningham <ncunningham@cyclades.com>
 *
 * This file is released under the GPLv2.
 *
 * It implements support for dynamically allocated bitmaps that are
 * used for temporary or infrequently used pageflags, in lieu of
 * bits in the struct page flags entry.
 */

#ifndef DYN_PAGEFLAGS_H
#define DYN_PAGEFLAGS_H

#include <linux/mm.h>

typedef unsigned long ** dyn_pageflags_t;

#define BITNUMBER(page) (page_to_pfn(page))

#define PAGEBIT(page) ((int) ((page_to_pfn(page))%(8 * sizeof(unsigned long))))

#define BITS_PER_PAGE (PAGE_SIZE * 8)
#define PAGES_PER_BITMAP ((max_mapnr + BITS_PER_PAGE - 1) / BITS_PER_PAGE)
#define PAGENUMBER(page) (BITNUMBER(page) / BITS_PER_PAGE)

#define PAGEINDEX(page) ((BITNUMBER(page) - (BITS_PER_PAGE * PAGENUMBER(page)))/(8*sizeof(unsigned long)))

#define PAGE_UL_PTR(bitmap, pagenum) ((bitmap[PAGENUMBER(pagenum)])+PAGEINDEX(pagenum))

#define __get_next_bit_on_safe(bitmap, counter) \
	do { \
		(counter)++; \
	} while(((counter) < max_mapnr) && \
		(!test_bit(PAGEBIT(pfn_to_page((counter))), \
			PAGE_UL_PTR(bitmap, pfn_to_page((counter))))));

static inline int __get_next_bit_on(dyn_pageflags_t bitmap, int counter)
{
	do {
		counter++;
	} while((counter < max_mapnr) &&
		(!test_bit(PAGEBIT(pfn_to_page(counter)),
			PAGE_UL_PTR(bitmap, pfn_to_page(counter)))));
	return counter;
}

/* With the above macros defined, you can do...

#define PageInUse(page)	\
	test_bit(PAGEBIT(page), PAGE_UL_PTR(in_use_map, page))
#define SetPageInUse(page) \
	set_bit(PAGEBIT(page), PAGE_UL_PTR(in_use_map, page))
#define ClearPageInUse(page) \
	clear_bit(PAGEBIT(page), PAGE_UL_PTR(in_use_map, page))
*/

extern void clear_dyn_pageflags(dyn_pageflags_t pagemap);
extern int allocate_dyn_pageflags(dyn_pageflags_t *pagemap);
extern int free_dyn_pageflags(dyn_pageflags_t *pagemap);

#endif
